"""
structure.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""

from .errata import pivot
from .trajectory import Trajectory

from molecular.analysis.protein.hydrophobicity import rmsd
from molecular.external import stride

from glovebox import GloveBox
import numpy as np
import os

# os.environ["MODIN_ENGINE"] = 'dask'
# import modin.pandas as pd
import pandas as pd


# Structure class
class Structure:
    """
    A `Structure` is an object that stores a single three-dimensional "structure"

    "Structure" can be used loosely here because it applies to proteins, proteins in water baths, proteins in lipid
    environments, ligands in water, etc.

    The `Structure` instance generally follows PDB format [1] in that in contains,
      - id
      - x
      - y
      - z
      - alpha
      - beta
    """

    # Required columns
    columns = [
        'atom_id',
        'atom',
        'residue',
        'chain',
        'residue_id',
        'x',
        'y',
        'z',
        'alpha',
        'beta',
        'segment',
        'element',
    ]

    # Initialize class instance
    def __init__(self, data=None, topology=None):
        """
        Initialize class instance
        """

        # TODO should this be protected?
        # Save data or create blank DataFrame
        self._data = data if data is not None else pd.DataFrame(columns=self.columns)

        # Construct topology
        self.topology = None

    # Add
    def __add__(self, other):
        """
        Add constant to Cartesian coordinates of Structure instance.

        Parameters
        ----------
        other

        Returns
        -------

        """

        for column in ['x', 'y', 'z']:
            self._data[column] += other

        return self.copy()

    # Get attribute
    def __getattr__(self, item):
        if item not in self._data:
            raise AttributeError('%s not in Structure' % item)

        return self._data[item].values

    # Get atom from Structure by index
    def __getitem__(self, item):
        """

        Parameters
        ----------
        item

        Returns
        -------

        """

        # Make sure that item a valid atom_id
        if item not in self._data['atom_id']:
            raise AttributeError('%s not a valid atom_id' % item)

        # Create copy of the row
        data = self._data[self._data['atom_id'] == item].copy()

        # Sanity
        if len(data) != 1:
            raise AttributeError('multiple atom_id %s found in Structure' % item)

        # Return (as Series)
        return data.iloc[0, :]

    # Length of structure (returns number of atoms)
    def __len__(self):
        """
        Number of atoms in the Structure instance.

        Returns
        -------
        int
            Number of atoms in the Structure instance.
        """

        return len(self._data)

    # Add new atoms from list
    def add_atoms(self, **kwargs):
        """

        Parameters
        ----------
        kwargs

        Returns
        -------
        """

        # _add_atoms(self.data, self.columns, kwargs)
        pass

    # Apply
    def apply(self, function):
        return function(self)

    # TODO is this too specific? Evaluate if Structure should be more generalized
    # Compute secondary structure
    def compute_secondary_structure(self, recompute=False, executable='stride'):
        # If secondary_structure is not in the Structure, or recompute is true, compute
        if 'secondary_structure' not in self._data or recompute:
            # TODO in lieu of directly porting stride to Python, how can this be sped up?
            # Write out PDB of structure to glovebox
            gb = GloveBox('molecular-stride', persist=True)
            temp_pdb = os.path.join(gb.path, 'temp.pdb')
            self.to_pdb(temp_pdb)

            # Run STRIDE
            secondary_structure = stride(temp_pdb, executable=executable)[['residue_id', 'secondary_structure']]

            # Clean glovebox
            gb.clean()

            # Drop secondary_structure if it's already in self._data
            self._data = self._data.drop(columns='secondary_structure', errors='ignore')

            # Merge STRIDE results with Structure to store
            self._data = self._data.merge(secondary_structure, how='inner', on='residue_id', validate='m:1')

        # Return
        return self._data[['residue_id', 'secondary_structure']].drop_duplicates()

    # Copy
    def copy(self):
        """
        Create a copy of Structure instance.

        Returns
        -------
        molecular.Structure
            Copy of instance.
        """

        return Structure(self._data.copy())

    # Pivot
    def pivot(self, index, columns=None, values=None, aggfunc='mean'):
        return pivot(self._data, index=index, columns=columns, values=values, aggfunc=aggfunc)

    # TODO should this return a Structure or Trajectory? TBD
    # TODO what if there were an intermediate object, like a Query class that knew how to handle subindexing / setting
    # Query
    def query(self, text):
        # TODO make this so it's customizable
        # Run macros
        text = text.lower().replace('peptide', 'residue in ["ALA", "ARG", "ASN", "ASP", "CYS", "GLN", "GLU", "GLY", ' +
                                    '"HIS", "HSD", "ILE", "LEU", "LYS", "MET", "PHE", "PRO", ' +
                                    '"SER", "THR", "TRP", "TYR", "VAL"]')

        # Get indices and parsed data
        indices = self._data.query(text).index.values
        data = self._data.loc[self._data.index.isin(indices), :]

        # If self is Trajectory, return a Trajectory
        if isinstance(self, Trajectory):
            result = Trajectory(data)
        else:
            result = Structure(data)

        # Return to user
        return result

    # TODO extend this to another trajecotry
    # Compute RMSD
    def rmsd(self, structure):
        """
        Compute root-mean-square deviation (RMSD) with another structure.

        Parameters
        ----------
        structure :

        Returns
        -------

        """

        return rmsd(self.xyz, structure.xyz)

    # Set topology
    def set_topology(self, topology):
        """
        Set topology

        Parameters
        ----------
        topology : Topology
            Instance of Topology class
        """

        self.topology = topology

    # Write to CSV
    def to_csv(self, path, topology=False):
        pass

    # Save as PDB
    def to_pdb(self, filename_or_buffer):
        """

        Parameters
        ----------
        filename_or_buffer

        Returns
        -------

        """

        # Create a copy of the data
        data = self._data.copy()

        # Add ATOM record
        data['record'] = 'ATOM'

        # TODO this should also be done for residue id probably
        # Change base-0 to base-1 for atom_id
        if data['atom_id'].min() == 0:
            data['atom_id'] += 1

        # Format atom
        i = data['atom'].str.len() == 1
        data.loc[i, 'atom'] = ' ' + data.loc[i, 'atom'] + '  '

        i = data['atom'].str.len() == 2
        data.loc[i, 'atom'] = ' ' + data.loc[i, 'atom'] + ' '

        # Select pertinent columns
        data = data[['record', 'atom_id', 'atom', 'residue', 'chain', 'residue_id',
                     'x', 'y', 'z', 'alpha', 'beta', 'segment', 'element']]

        # Write out PDB file
        np.savetxt(
            filename_or_buffer,
            data,
            fmt='%-6s%5i %4s%4s%2s%4i%12.3f%8.3f%8.3f%6.2f%6.2f%9s%2s',
            header='CRYST1    0.000    0.000    0.000  90.00  90.00  90.00 P 1           1',
            footer='END',
            comments=''
        )

    # Convert to trajectory
    def to_trajectory(self, structure_id=None):
        # Create and/or update structure_id if necessary
        if 'structure_id' not in self._data:
            structure_id = 0
        if structure_id is not None:
            self._data['structure_id'] = structure_id

        # Return Trajectory
        return Trajectory(self._data)

    # Get coordinates
    @property
    def xyz(self):
        """

        Returns
        -------

        """

        return self._data[['x', 'y', 'z']].values


# Trajectory cannot simply be a list of Structures
# class Trajectory(Structure):
#     """
#
#     """
#
#     # Initialize instance of Trajectory class
#     # noinspection PyMissingConstructor
#     def __init__(self, data=None):
#         if isinstance(data, pd.DataFrame):
#             self._data = data
#
#     # Get Structure from Trajectory
#     def __getitem__(self, item):
#         """
#
#
#         Parameters
#         ----------
#         item
#
#         Returns
#         -------
#
#         """
#
#         # Convert item to integer array if not already
#         item = np.array(item, dtype='int').reshape(-1)
#         if len(item) not in [1, 2]:
#             raise AttributeError('index must be structure_id or list with [structure_id, atom_id]')
#
#         # First, get structure
#         result = self.get_structure(item[0])
#
#         # If there's a second dimension, get an atom
#         if len(item) == 2:
#             result = result[item[1]]
#
#         # Return
#         return result
#
#     # Get number of structures in Trajectory
#     def __len__(self):
#         """
#         Number of structures in Trajectory.
#
#         """
#
#         return self._data['structure_id'].nunique()
#
#     # Add a structure
#     # noinspection PyProtectedMember
#     def add_structure(self, structure):
#         structure._data['structure_id'] = self._data['structure_id'].max() + 1
#         self._data = pd.concat([self._data, structure._data])
#
#     # TODO this can be parallelized
#     # Apply
#     def apply(self, function):
#         result = None
#         for structure_id in self._data['structure_id'].unique():
#             # Run the apply function from the Structure with structure_id
#             result_ = self.get_structure(structure_id).apply(function)
#
#             # If the result is a series, convert to frame and use structure_id as the index
#             if isinstance(result_, pd.Series):
#                 result_ = result_.to_frame().T
#                 result_.index = [structure_id]
#                 result_.index.name = 'structure_id'
#
#             if result is None:
#                 result = result_
#             else:
#                 if isinstance(result, pd.DataFrame):
#                     result = pd.concat([result, result_])
#                 elif isinstance(result, list):
#                     result.append(result_)
#                 else:
#                     result = [result, result_]
#
#         return result
#
#     # TODO this can be parallelized
#     # Compute secondary structure
#     def compute_secondary_structure(self, recompute=False, executable='stride'):
#         for structure_id in self._data['structure_id'].unique():
#             self.get_structure(structure_id).compute_secondary_structure(recompute=recompute, executable=executable)
#
#     # Get structure
#     def get_structure(self, structure_id):
#         # Make sure structure_id is an array
#         structure_id = np.array(structure_id, dtype='int').reshape(-1)
#
#         # Parse DataFrame for appropriate rows and return as Structure
#         return Structure(self._data.loc[self._data['structure_id'].isin(structure_id), :])



"""
quantity.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""

from abc import ABCMeta
import numpy as np
import pandas as pd


class Ensemble(metaclass=ABCMeta):
    def __init__(self):
        pass


# noinspection PyPep8Naming
class NPT(Ensemble):
    def __init__(self, N, P, T):
        self.N = N
        self.P = P
        self.T = T
        Ensemble.__init__(self)


# noinspection PyPep8Naming
class NVT(Ensemble):
    def __init__(self, N, V, T):
        self.N = N
        self.V = V
        self.T = T
        Ensemble.__init__(self)


# TODO a Quantity can have an energy attached to it? That would make this a reaction coordinate. Maybe a RC is child?
# TODO a Quantity is a dependent of an Ensemble?
class Quantity:
    """
    A `Quantity` is something that has measures something (along the wide axis) versus observations (along the long
    axis).

    Because this is a proof of concept, `Quantity` is limited to two-dimensions. Higher-dimensional structures can still
    be considered as a flattened list. For instance, if you wanted to look at secondary structure per residue per
    observation, you would have to construct your wide axis as, "residue0_H, residue0_T, residue0_C, ..."

    The benefit of a `Quantity` is that it can be block-averaged, used in WHAM ...

    Examples
    --------
    1. Energy vs observations
    2. Number of residues in helical state vs observations
    3. Flag if each residue is helix vs observations
    """

    # Initialize instance of Quantity class
    def __init__(self, data, ensemble=None):
        """

        Parameters
        ----------
        data
        """

        self._data = data
        self._ensemble = ensemble

    # The length of the Quantity is equal to the number of observations
    def __len__(self):
        """
        Number of observations for the Quantity instance.

        Returns
        -------
        int
            Number of observations
        """

        # Return number of observations
        return len(self._data)

    # Represent Quantity by self._data
    def __repr__(self):
        return pd.DataFrame.__repr__(self._data)

    # Assign blocks
    def _assign_blocks(self, num_blocks, block_size):
        """

        Parameters
        ----------
        num_blocks
        block_size

        Returns
        -------

        """

        # Observations
        num_observations = len(self)
        observations = np.arange(num_observations)

        # If block_size is set, set number of blocks by dividing number of observations by block size
        if block_size is not None:
            num_blocks = np.floor(num_observations / block_size)

        # Else if num_blocks is None, set to square root of num_observations
        elif num_blocks is None:
            num_blocks = np.floor(np.sqrt(num_observations))

        # If num_observations <= num_blocks, do not move forward
        if num_observations <= num_blocks:
            raise AttributeError('num_observations cannot be <= num_blocks')

        # Create bins
        bins = np.floor(np.linspace(start=observations[0], stop=observations[-1], num=num_blocks + 1))
        bins[-1] = np.inf

        # Return blocks
        return np.digitize(observations, bins=bins, right=False) - 1

    # Compute block averages
    def block_averages(self, num_blocks=None, block_size=None, blocks=None):
        # Determine blocks
        if blocks is None:
            blocks = self._assign_blocks(num_blocks, block_size)

        # Assign block
        self._data['block'] = blocks

        # For every block, compute the average
        return self._data.astype(float).pivot_table(index='block', values=self.columns, aggfunc='mean')

    # Compute the block error
    def block_error(self, num_blocks=10, block_size=None, blocks=None):
        # Determine blocks
        if blocks is None:
            blocks = self._determine_blocks(num_blocks, block_size)
        num_blocks = len(blocks)

        # Compute block averages
        block_averages = self.block_average(num_blocks=num_blocks, block_size=block_size)

        # Compute block average of the squared quantity
        quantity_squared = Quantity(np.square(self._data))
        block_averages_squared = quantity_squared.block_average(num_blocks=num_blocks, block_size=block_size)

        # Compute the standard error
        return np.sqrt((block_averages_squared - np.square(block_averages)) / (num_blocks - 1))

    # Columns of Quantity
    @property
    def columns(self):
        return self._data.columns

    # TODO should there be an axis = None here?
    def mean(self, axis=1):
        """

        Parameters
        ----------
        axis

        Returns
        -------

        """

        return self._data.mean(axis=axis)

    def std(self, axis=None):
        pass

    def sum(self, axis=None):
        pass



"""
transform.py

language: Python
version: 3.x
author: C. Lockhart <chris@lockhartlab.org>
"""

import logging
from numba import njit
import numpy as np
import pandas as pd

# Get the molecular.transform logger
logger = logging.getLogger('molecular.transform')


# Compute the center of the Trajectory
def center(a, weights=None):
    r"""
    Compute the center `m` of Trajectory `a`.

    .. math :: m = \frac{\sum_{i=1}^N w_i (x_i + y_i + z_i)} {\sum_{i=1}^N w_i}

    Here, :math:`w_i` refers to weights applied to individual atoms :math:`i`.

    Parameters
    ----------
    a : molecular.Trajectory
    weights : array-like
        (Optional) Weights for atoms in the trajectory. Follows the definition from :ref:`numpy.average`.

    Returns
    -------
    pandas.DataFrame
        Center of each structure in the Trajectory.
    """

    # Compute center
    m = a.coordinates.groupby('structure_id').mean()

    # Update log
    logging.info(f'computed center of {a.designator}')

    # Return
    return m


def fit(a, b):
    """
    Fit Trajectory `b` to Trajectory `a` in a pairwise fashion.

    Parameters
    ----------
    a : molecular.Trajectory
    b : molecular.Trajectory

    Returns
    -------

    """

    # Number of structures
    n_structures = a.n_structures
    if n_structures != b.n_structures:
        raise AttributeError('a and b must have the same number of structures')

    # Number of atoms
    n_atoms = a.n_atoms
    if n_atoms != b.n_atoms:
        raise AttributeError('a and b must have the same number of atoms')

    # Number of dimensions
    n_dim = 3

    # Move structures to center
    a_xyz_center = center(a)
    b_xyz_center = center(b)
    a_xyz = a.xyz - a_xyz_center
    b_xyz = b.xyz - b_xyz_center

    # Transpose b
    b_xyz_transpose = np.transpose(b_xyz, axes=[0, 2, 1])

    # Compute covariance matrix
    covariance_matrix = np.matmul(b_xyz_transpose, a_xyz)

    # Get rotation matrix
    rotation_matrix = _get_rotation_matrix(covariance_matrix)

    # Perform optimal rotation
    b_xyz = np.matmul(b_xyz, rotation_matrix)

    # Move `b` to `a` center
    b_xyz = b_xyz + a.center()

    # Return
    return b_xyz


# Center Trajectory
# def center(a, weights=None, inplace=False):
#     r"""
#     Compute the center of Trajectory `a`.
#
#     .. math :: center = \frac{1}{N} \Epsilon w_i (x_i + y_i + z_i)
#
#     Parameters
#     ----------
#     a : molecular.Trajectory
#     weights : array-like
#         (Optional) Weights for atoms in the trajectory. Follows the definition from :ref:`numpy.average`.
#     inplace : bool
#         Should the Trajectory be centered in-place? (Default: False)
#
#     Returns
#     -------
#     molecular.Trajectory
#         Trajectory centered at origin.
#     """
#
#     # Compute center
#     xyz = np.average(a.xyz, weights=weights, axis=1, returned=False)
#
#     # Update log
#     logging.info(f'computed center of {a.designator}')
#
#     # Copy if not an inplace transformation
#     if not inplace:
#         a = a.copy()
#
#     # Update the coordinates
#     a.xyz = xyz
#
#     # Return the copy if necessary
#     if not inplace:
#         return a
#


# Move Trajectory
# noinspection PyProtectedMember
def move(a, by=None, to=None, inplace=True):
    r"""
    Move Trajectory `a` either `by` a constant or `to` a location. The two arguments `by` and `to` cannot be set
    simultaneously.

    Parameters
    ----------
    a : molecular.Trajectory
    by : array-like
        (Optional) Move `a` by a constant.
    to : array-like
        (Optional) Move `a` to a location.
    inplace : bool
        Should we perform the translation on a copy of the Trajectory (Default: False)

    Returns
    -------
    None or molecular.Trajectory
        If `return_copy` is set, we return the moved Trajectory.
    """

    # If both `by` and `to` are None (or both set), we don't know what to do
    if (by is None and to is None) or (by is not None and to is not None):
        raise AttributeError('either `by` or `to` must be set')

    # If `to` is set, let's refactor this as `by`
    if to is not None:
        m = center(a)
        by = to - m  # xyz - center + to is the idea

    # `by` must be array for ease
    if not isinstance(by, (np.ndarray, pd.DataFrame)):
        # noinspection PyTypeChecker
        by = np.array(by)

    # Should we make a copy?
    if not inplace:
        a = a.copy()

    # Perform translation
    # TODO need to come up with robust tests about when array links are preserved
    a.coordinates = a.coordinates + by

    # Update log
    logging.info(f'translated {a.designator}')

    # Return if necessary
    if not inplace:
        return a


def _fit(a_xyz, b_xyz, backend='python'):
    """
    `a_xyz` and `b_xyz` must be paired.

    Parameters
    ----------
    a_xyz : numpy.ndarray
    b_xyz : numpy.ndarray
    backend : str

    Returns
    -------

    """

    # Extract dimensions
    n_structures, n_atoms, n_dim = a_xyz.shape

    # Sanity check
    if n_structures != b_xyz.shape[0]:
        raise AttributeError('a_xyz and b_xyz must have the same number of structures')
    if n_atoms != b_xyz.shape[1]:
        raise AttributeError('a_xyz and b_xyz must have the same number of atoms')

    # Centers
    a_xyz_center = np.tile(a_xyz.mean(axis=1), n_atoms).reshape(n_structures, n_atoms, n_dim)
    b_xyz_center = np.tile(b_xyz.mean(axis=1), n_atoms).reshape(n_structures, n_atoms, n_dim)

    # Move structures to center
    a_xyz = a_xyz - a_xyz_center
    b_xyz = b_xyz - b_xyz_center

    # Transpose b
    b_xyz_transpose = np.transpose(b_xyz, axes=[0, 2, 1])

    # Compute covariance matrix
    covariance_matrix = np.matmul(b_xyz_transpose, a_xyz)

    # Get rotation matrix
    rotation_matrix = _get_rotation_matrix(covariance_matrix)

    # Perform optimal rotation
    b_xyz = np.matmul(b_xyz, rotation_matrix)

    # Move b to a center
    b_xyz = b_xyz + a_xyz_center

    # Return
    return b_xyz


@njit
def _get_rotation_matrix(covariance_matrix):
    """
    Compute the rotation matrix from the method from Kabsch (1976) Acta Cryst. A32.



    Parameters
    ----------
    covariance_matrix

    Returns
    -------

    """
    n_structures = covariance_matrix.shape[0]
    rotation_matrix = []
    for i in range(n_structures):
        u, s, v = np.linalg.svd(covariance_matrix[i, :, :], full_matrices=True)
        s = np.diag(np.array([1., 1., np.sign(np.linalg.det(np.dot(u, v)))]))
        r = np.dot(np.dot(u, s), v)
        rotation_matrix.append(r)
    return rotation_matrix


if __name__ == '__main__':
    import molecular as mol

    trj = mol.read_pdb('../tests/samples/trajectory.pdb')
    move(trj, to=(0, 0, 0))
    assert (trj.center().to_numpy() < 1e-7).max()

    # TODO add this to unit tests
    sel = trj.select(segment='GB1')
    print(center(sel))
    move(sel, by=(5, 5, 5))
    print(center(sel))
    print(center(trj))

    # TODO add something like this to unit tests
    # small = trj.get_structure(0).select(atom_id=[0, 1]).copy()
    # xyz = np.array(small.xyz)
    # # sel = small.select(atom_id=0)
    # # sel = mol.Trajectory(small.get_atoms(np.array(0)))
    # sel = mol.Trajectory(small._xyz[slice(None), [0], slice(None)])  # when you slice by a list, you always get a copy
    # sel._xyz[:] = sel._xyz[:] + 5
    # print(np.max(small.xyz - xyz))



"""
Functions related to time series analysis.
"""

from numba import njit
import numpy as np


# noinspection PyShadowingNames
def acorr(a, fft=False):
    r"""
    Compute the lag-:math:`k` autocorrelation :math:`\rho(k)` from dataset :math:`a`.

    .. math :: \rho(k) = \frac{\gamma(k)}{\gamma(0)}

    The function :math:`\gamma` is from :func:`acov`.

    Parameters
    ----------
    a : numpy.ndarray
        One-dimensional array.
    fft : bool
        Should fast Fourier transform be used?

    Returns
    -------
    numpy.ndarray
        Autocorrelation function
    """

    # Compute autocovariance
    gamma = acov(a, fft=fft)
    return gamma / gamma[0]


# noinspection DuplicatedCode,PyShadowingNames
def acov(a, fft=False):
    r"""
    Compute the unbiased auto-covariance function :math:`\gamma(k)` from dataset :math:`a` with :math:`N` samples for
    lag-time :math:`k`.

    .. math :: \gamma(k)=\frac{1}{N-k}\sum_{t=1}^{N-k}(a_t - \mu)(a_{t+k} - \mu)

    Here, we also compute the mean

    .. math :: \mu=\frac{1}{N}\sum_t^Na_t

    Note: by default, all :math:`k` from 0 to N-1 are evaluated. Sampling deteriorates rapidly as :math:`k` increases.
    There is also a *biased* estimator for the autocovariance, which changes the denominator from :math:`n-k` to
    :math:`n` and has an effect of reducing the fluctuations due to error at large :math:`k`. To compute this, see
    :func:`statsmodels.tsa.stattools.acovf`.

    Parameters
    ----------
    a : numpy.ndarray
        1D array.
    fft : bool
        Should Fast-Fourier Transform be used?

    Returns
    -------
    numpy.ndarray
        Autocovariance function

    References
    ----------
    https://www.itl.nist.gov/div898/handbook/eda/section3/autocopl.htm
    """

    # We only want to compute this for 1D arrays
    if a.ndim != 1:
        raise AttributeError('must be 1D')

    # Remove the mean from the observations
    a -= np.mean(a)

    # Compute autocovariance gamma
    if not fft:
        # Compute a * a for all lag times using correlate and the unbiased autocovariance
        len_a = len(a)
        gamma = np.correlate(a, a, mode='full')[(len_a - 1):] / (len_a - np.arange(len_a))

    else:
        # Use statsmodels to compute FFT
        # noinspection PyPackageRequirements
        from statsmodels.tsa.stattools import acovf
        gamma = acovf(a, adjusted=True, demean=False, fft=True, missing='none', nlag=None)

    # Return
    return gamma


# noinspection PyShadowingNames
def inefficiency(a):
    """
    Compute the statistical inefficiency :math:`g` from the equilibration time :math:`\tau_{corr}`.

    .. math :: g = 1 + 2 \tau_{corr}.

    Parameters
    ----------
    a : numpy.ndarray

    Returns
    -------
    float
        Statistical inefficiency.
    """

    return 1. + 2. * tcorr(a)


# Estimate the standard error from the correlation time
# noinspection PyShadowingNames
def sem_tcorr(a, tol=1e-3):
    """
    Estimate of standard error of the mean derived from the correlation time.

    The main assumption is that sqrt(N) ~ sqrt(len(a) / tcorr(a))

    Should only be used for continuous time series data, e.g., from molecular dynamics. Discontinuous trajectories as
    produced by replica exchange or Monte Carlo are not applicable.

    Parameters
    ----------
    a : numpy.ndarray
    tol : float

    Returns
    -------

    """

    return np.std(a) * np.sqrt(tcorr(a) / len(a))


# TODO derive this cleanly
# noinspection PyShadowingNames
def tcorr(a, fft=False):
    r"""
    Compute correlation time :math:`\tau_{corr}` based on the autocorrelation function :math:`\rho(k)`.

    .. math :: \tau_{corr} = \sum_{t=1}^{T} (1 - \frac{t}{T}) \rho_{t}

    Note: only :math:`k` before the first occurrence :math:`\rho(k) = 0` are taken. Estimates after this are
    statistically unreliable.

    Parameters
    ----------
    a : numpy.ndarray
        1D array.
    fft : bool
        Should fast Fourier transform be used?

    Returns
    -------
    float
        Correlation time.

    References
    ----------

    """

    # Compute autocorrelation, reduce to observations before first < 0
    rho_ = acorr(a, fft=fft)[1:]
    rho = rho_[:np.min(np.where(rho_ < 0))]

    t_max = len(a)
    t = np.arange(1, len(rho) + 1)

    return np.sum((1. - t / t_max) * rho)


# noinspection PyShadowingNames
def _acorr(a):
    gamma = _acov(a)
    return gamma / gamma[0]


def _acorr_test(a, decimal=7, plot=True):
    rho0 = acorr(a)
    rho1 = _acorr(a)
    if plot:
        import matplotlib.pyplot as plt
        plt.figure()
        x = np.arange(len(rho0))
        plt.plot(x, rho0)
        plt.plot(x, rho1)
        plt.show()
    np.testing.assert_almost_equal(rho0, rho1, decimal=decimal)


@njit
def _acov(a):
    n = len(a)
    u = np.mean(a)
    gamma = np.zeros(n)
    for k in range(n):
        gamma[k] = np.mean((a[:n - k] - u) * (a[k:] - u))
    return gamma


def _acov_test(a):
    import time
    start_time = time.time()
    gamma0 = acov(a)
    end_time = time.time()
    print('acov={}'.format(end_time - start_time))

    start_time = time.time()
    gamma1 = acov(a, fft=True)
    end_time = time.time()
    print('acov={}'.format(end_time - start_time))

    start_time = time.time()
    gamma2 = _acov(a)
    end_time = time.time()
    print('acov={}'.format(end_time - start_time))

    np.testing.assert_almost_equal(gamma0, gamma1)
    np.testing.assert_almost_equal(gamma0, gamma2)


if __name__ == '__main__':
    n = 100000
    a = np.random.normal(loc=0, scale=10, size=n)
    print(tcorr(a))
    # _acov_test(a)
    # _acorr_test(a, decimal=5)
    # _tcorr_test(a)

    # print(np.std(a))
    # print(np.sqrt(np.cov(a)))
    # print(np.sqrt(xx - x*x))
    # print(correlation_time(a))
    # print(sem_tcorr(a))

    # import time
    # start_time = time.time()
    # rho = acorr(a)
    # end_time = time.time()
    # print(end_time - start_time)
    # start_time = time.time()
    # tau = tcorr(a)
    # end_time = time.time()
    # print(end_time - start_time)



"""
namd.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""

import os
import subprocess


# Class to run NAMD simulations
# TODO use glovebox to store all NAMD PIDs
class NAMD:
    """
    NAMD class, which launches simulations. Eventually this will do more error
    checking; however, for today it's bare minimum. This will break if there
    are duplicate parameters in the config file.

    >>> from molecular import namd
    >>> md = namd()
    >>> md.start()

    See Also
    --------
    https://github.com/radakb/pynamd
    """

    # Initialize class instance
    def __init__(self, config_file, executable='namd2', n_jobs=None, background=False):
        """
        Initialize NAMD run.

        Parameters
        ----------
        config_file : str
            Path to NAMD configuration file.
        executable : str
            Location of NAMD executable. (Default: `namd2` in current working directory).
        n_jobs : int
            Number of jobs.
        background : bool
            Flag to indicate if NAMD task should be executed in the background.
        """

        # Properties
        self.config_file = str(config_file)
        self.executable = str(executable)
        self.n_jobs = int(n_jobs) if n_jobs is not None else None
        self.background = bool(background)

        # Runtime functions
        self._process = None

    # Check if running
    def _check_if_running(self):
        # If the job is already running, we cannot start again
        if not self.poll():
            raise RuntimeError('NAMD already running')

    # Copy NAMD configuration
    def copy(self):
        """
        Copy NAMD configuration.

        Returns
        -------
        molecular.external.namd
            New instance of NAMD class
        """

        return NAMD(
            config_file=self.config_file,
            executable=self.executable,
            n_jobs=self.n_jobs,
            background=self.background
        )

    # PID
    @property
    def pid(self):
        """
        Return process ID.

        Returns
        -------
        int
            Process ID on system.
        """

        if self._process is None:
            result = None
        else:
            result = self._process.pid

        return result

    # Poll
    def poll(self):
        """
        Check if simulations are still running.

        Returns
        -------
        None or int
            None if simulations are running, return code if simulations are finished.
        """

        if self._process is None:
            result = None
        else:
            result = self._process.poll()

        return result

    # Start simulations
    def start(self, output_file=None):
        """
        Start NAMD simulations.

        Parameters
        ----------
        output_file : str
            Path to output file to write. If not specified, uses name of configuration file with extension '.out'.
        """

        # Check if NAMD is already running
        if self.poll():
            raise RuntimeError('NAMD already running')

        # Set an output file
        if output_file is None:
            output_file = os.path.splitext(self.config_file)[0] + '.out'

        # Set up command
        command = [self.executable]
        if self.n_jobs is not None:
            command += ['+p' + str(self.n_jobs)]
        command += [self.config_file]

        # Open output file and run
        with open(output_file, 'w') as buffer:
            self._process = subprocess.Popen(command, stdout=buffer)

        # Should we wait?
        if not self.background:
            self._process.wait()
            if self._process.poll() != 0:
                raise RuntimeError('NAMD job did not finish successfully')

    # Stop simulations
    def stop(self):
        # Check if NAMD is already running
        if self.poll():
            self._process.kill()

        # Otherwise, alert that NAMD is not running
        # TODO should this be warning?
        else:
            pass


# TODO https://stackoverflow.com/questions/39980323/are-dictionaries-ordered-in-python-3-6
class NAMDConfiguration:
    def __init__(self, config_file=None):
        self.parameters = {}
        if config_file is not None:
            self.read(config_file)

    def __repr__(self):
        return str(self.parameters)

    def add(self, parameter, *args):
        args = list(args)
        # if len(args) > 1:
        #     args = [args]
        args = [args]
        self.parameters[parameter] = self.parameters.get(parameter, []) + args

    def read(self, config_file):
        with open(config_file, 'r') as buffer:
            lines = buffer.readlines()
            for line in lines:
                # Parse up to comment
                if '#' in line:
                    line = line[:line.find('#')]

                # Strip line
                line = line.strip()

                # If line is not empty,
                if len(line) != 0:
                    # Add
                    self.add(*line.split())

    def write(self, config_file):
        pass



    # def read_config(self, filename):
    #     with open(filename, 'r') as buffer:
    #         self.parameters = buffer.read()
    #
    # def read_output(self):
    #     return read_NAMD_output(self.name+'.out')
    #
    # def write_config(self):
    #     with open(self.name + '.namd', 'w') as buffer:
    #         buffer.write(self.parameters)
#
#
# # Function for reading NAMD output
# def read_NAMD_output(fname,ignore_first=False):
#     file=open(str(fname),'r')
#     lines=file.read().split('\n')
#     file.close()
#     keys=None
#     values=[]
#     for line in lines:
#         if line[:6] == 'ETITLE':
#             keys_=line.split()[1:]
#             if keys is None: keys=keys_
#             if keys != keys_:
#                 raise ValueError('keys change')
#         elif line[:6] == 'ENERGY':
#             values_=line.split()[1:]
#             values.append(values_)
#     if ignore_first: values=values[1:]
#     values=np.array(values,dtype=float)
#     df=pd.DataFrame()
#     for i,key in enumerate(keys):
#         df[key]=pd.Series(values[:,i])
#     return df



"""
sequence.py

>>> abeta = Protein('YEVHHQKLVFFAEDVGSNKGAIIGLMVGGVV')
"""

from typelike import ArrayLike


class Sequence:
    def __init__(self):
        pass


class Protein(Sequence):
    """
    Construct for dealing with protein sequence.

    """

    def __init__(self, sequence, ):
        """


        Parameters
        ----------
        sequence : np.ndarray
            List of residues.
        """

        # Convert to list if not already list
        super().__init__()
        if not isinstance(sequence, ArrayLike):
            residues = list(sequence)

        # Which format are these residues in?
        len_sequence = len(sequence[0])
        if len_sequence == 1:
            residues = _letter_to_code(sequence)
        elif len_sequence != 3:
            raise AttributeError('must supply residue triplets')

        # Save
        self.sequence = sequence

    def to_letters(self, join=False):
        code_to_letter = {
            'ALA': 'A',
            'ARG': 'R',
            'ASN': 'N',
            'ASP': 'D',
            'CYS': 'C',
            'GLN': 'Q',
            'GLU': 'E',
            'GLY': 'G',
            'HIS': 'H',
            'HSD': 'H',
            'ILE': 'I',
            'LEU': 'L',
            'LYS': 'K',
            'MET': 'M',
            'PHE': 'F',
            'PRO': 'P',
            'SER': 'S',
            'THR': 'T',
            'TRP': 'W',
            'TYR': 'Y',
            'VAL': 'V',
        }

        sequence = [code_to_letter[code] for code in self.sequence]

        if join:
            sequence = ''.join(sequence)

        return sequence


    def to_str(self):
        pass


def _letter_to_code(residues):
    # Make sure we're in the right format
    # noinspection DuplicatedCode
    letter_to_code = {
        'A': 'ALA',
        'R': 'ARG',
        'N': 'ASN',
        'D': 'ASP',
        'C': 'CYS',
        'Q': 'GLN',
        'E': 'GLU',
        'G': 'GLY',
        'H': 'HIS',
        'I': 'ILE',
        'L': 'LEU',
        'K': 'LYS',
        'M': 'MET',
        'F': 'PHE',
        'P': 'PRO',
        'S': 'SER',
        'T': 'THR',
        'W': 'TRP',
        'Y': 'TYR',
        'V': 'VAL'
    }
    return [letter_to_code.get(residue, residue) for residue in residues]


"""
protein.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""

# from molecular import include_dir
from molecular.geometry import distance
from typelike import ListLike

import numpy as np
import os
import pandas as pd
import sys

# TODO move this to root
include_dir = os.path.abspath(__file__ + '/../../include')

# Reference to this module
this = sys.modules[__name__]

# Cache to store data in session
this.cache = {}


# Hydrophobicity scale class
class HydrophobicityScale:
    """
    Store a hydrophobicity scale that can be used for analysis
    """

    # Initialize instance
    def __init__(self, scale=None):
        """
        Initialize instance of ``HydrophobicityScale`` class
        """

        # Initialize data elements
        self.data = None
        self.residue_name_format = None

        # If scale is not None, load
        if scale is not None:
            self.load(scale)

    # Parse the scale
    def __getitem__(self, item):
        # If integer, assume that we're referring to the index
        if isinstance(item, int):
            element = self.data.iloc[item]

        # Otherwise, use the index name
        else:
            element = self.data[item]

        # Return
        return element

    # Load scale
    def load(self, scale):
        """
        Load scale by name

        Parameters
        ----------
        scale : str
            Name of hydrophobicity scale (Default: 'wimley-white')
        """

        # Convert scale to lowercase, convert - to _
        # TODO enable parsing of part of the scale name
        scale = scale.lower().replace('-', '_')

        # Based on scale, get filename
        filename = os.path.join(include_dir, 'protein', 'hydrophobicity_scales', scale + '.csv')

        # Read
        self.data = pd.read_csv(filename).set_index('residue').loc[:, 'hydrophobicity']

        # Get residue name format
        self.residue_name_format = _get_residue_name_format(self.data.index.values[0])


# Convert residue name format
def _convert_residue_name_format(residue, old_format, new_format, squeeze=True):
    """
    Convert ``residue`` from ``old_format`` to ``new_format``

    Parameters
    ----------
    residue : str or ListLike of str
    old_format : str
    new_format : str

    Returns
    -------

    """

    # Make sure residue is list
    if not isinstance(residue, ListLike):
        residue = [residue]

    # Convert residue list to pandas DataFrame
    df = pd.DataFrame({'old_name': residue})

    # Load RCT; filter RCT for old_format and new_format; create local copy; drop columns
    rct = _load_residue_conversions_table()
    x = rct['old_format'] == old_format
    y = rct['new_format'] == new_format
    rct = rct[x & y].copy().drop(columns=['old_format', 'new_format'])

    # Merge df with RCT to get new name
    df = df.merge(rct, how='left', on='old_name', validate='m:1')

    # Return new name
    return df['new_name'].values.squeeze()


# Get the format of the residue name based on length
def _get_residue_name_format(residue):
    # First off, make sure residue is a string, otherwise this will never work
    if not isinstance(residue, str):
        raise AttributeError('residue must be str; received {}'.format(residue))

    # Second, based on the length, assign the format
    n = len(residue)
    if n == 1:
        residue_format = 'letter'
    elif n == 3:
        residue_format = 'abbreviation'
    else:
        residue_format = 'full'

    # Return
    return residue_format


# Helper function to load hydrophobicity scale from cache or file
def _load_hydrophobicity_scale(scale):
    # Key
    key = 'hydrophobicity_scale_' + scale

    # If scale not in cache, add it
    if key not in this.cache:
        this.cache[key] = HydrophobicityScale(scale)

    # Return
    return this.cache[key]


# Helper function to load residue conversions table from cache / file
def _load_residue_conversions_table():
    # If residue_conversions_table has not yet been loaded, load
    if 'residue_conversions_table' not in this.cache:
        # Get filename
        filename = os.path.join(include_dir, 'protein', 'residue_conversions.csv')

        # Read
        this.cache['residue_conversions_table'] = pd.read_csv(filename)

    # Return the table
    return this.cache['residue_conversions_table'].copy()


# Compute hydrophobic moment from sequence
def compute_hydrophobic_moment(sequence, offset=100., scale='wimley-white'):
    """
    Compute the hydrophobic moment from a sequence

    Parameters
    ----------
    sequence : str or list of str
        Amino acid sequence
    offset : float
        Angle (in degrees) between amino acids (Default: 100 degrees)
    scale : str
        Hydrophobicity scale to use (Default: 'wimley-white')

    Returns
    -------
    float
        Hydrophobic moment magnitude
    """

    # Coerce sequence into correct form
    if isinstance(sequence, str):
        sequence = list(sequence)

    # Convert to numpy array for convenience
    sequence = np.array(sequence)

    # Load the right hydrophobicity scale
    if isinstance(scale, str):
        scale = _load_hydrophobicity_scale(scale)

    # At this pint, scale must by an instance of HydrophobicScale
    if not isinstance(scale, HydrophobicityScale):
        raise AttributeError('scale must be a known hydrophobicity scale or HydrophobicScale object')

    # Get the format of sequence entries and scale format
    sequence_format = _get_residue_name_format(sequence[0])

    # Convert the sequence format to the same as scale format
    sequence = _convert_residue_name_format(sequence, sequence_format, scale.residue_name_format)

    # Number of amino acid residues
    n_residues = len(sequence)

    # Convert the offset to radians, flip the sign, and compute angles between residues
    offset = -np.deg2rad(offset)
    angles = np.arange(0, n_residues * offset, offset)

    # Every residue has a 2D hydrophobic vector (-y points toward the hydrophobic medium)
    vector = np.zeros((n_residues, 2))
    vector[:, 0] = scale[sequence] * np.cos(angles)
    vector[:, 1] = scale[sequence] * np.sin(angles)

    # Return the hydrophobic moment
    return distance(vector, method='euclidean')


# RMSD
def rmsd(coord1, coord2, align=True):
    """
    Compute the RMSD between two sets of coordinates

    Parameters
    ----------
    coord1 : ArrayLike
    coord2 : ArrayLike

    Returns
    -------

    """

    # Sanity
    if not coord1.shape == coord2.shape:
        raise AttributeError('coordinates must be same shape')

    # Return RMSD
    return np.linalg.norm(coord1 - coord2)/np.sqrt(coord1.shape[0])








"""
replica_exchange.py

author: C. Lockhart <chris@lockhartlab.org>
language: Python3
"""


import numpy as np


# Create a temperature schedule
def temp_schedule(temp_min=300, temp_max=440, n_temps=40, mode='geometric'):
    r"""
    Create a temperature schedule that could be used, for instance, with replica exchange.

    There are several choices for `mode`. Note that :math:`T` refers to the temperature at :math:`i = 1 ... R`, where
    :math:`T_1` is `temp_min` and :math:`T_R` is `temp_max`. In total, there are :math:`R` temperatures (= `n_temps`).

    * "geometric" [#]_

    .. math :: T_i = T_1 \left( \frac{T_R}{T_1} \right)^{\frac{i-1}{R-1}}

    * "linear"

    .. math :: T_i = T_1 + (i-1) \frac{T_R-T_1}{R-1}

    * "parabolic" (Note if `n_temps` is even, `temp_max` won't directly be sampled).

    .. math :: T_i = T_1 - \frac{T_R-T_1}{\left( \frac{R-1}{2} \right) ^2} (i-1) (i-R)

    Parameters
    ----------
    temp_min : float
        Lowest temperature
    temp_max : float
        Highest temperature
    n_temps : int
        Number of temperatures
    mode : str
        Mode to produce schedule. Valid options include "geometric", "linear", "parabolic". Any substring will match,
        but the preference should be to use the full option label. (Default: "geometric")

    Returns
    -------
    numpy.ndarray
        Temperature schedule

    Examples
    --------
    .. plot::
       :include-source:

       import matplotlib.pyplot as plt
       import molecular as mol

       n_temps = 10
       geometric = mol.temp_schedule(300, 440, n_temps, 'geometric')
       linear = mol.temp_schedule(300, 440, n_temps, 'linear')
       parabolic = mol.temp_schedule(300, 440, n_temps, 'parabolic')

       plt.figure()
       plt.plot(range(n_temps), geometric, label='geometric')
       plt.plot(range(n_temps), linear, label='linear')
       plt.plot(range(n_temps), parabolic, label='parabolic')
       plt.xlabel('index')
       plt.ylabel('temperature')
       plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
       plt.show()

    References
    ----------
    .. [#] Nymeyer, H., Gnanakaran, S., & García, A. E. (2004) Atomistic simulations of protein folding, using the
       replica exchange algorithm. *Methods Enzymol.* **383**: 119-149.
    """

    mode = mode.lower()

    if mode in 'geometric':
        schedule = temp_min * np.power(temp_max / temp_min, np.arange(n_temps) / (n_temps - 1.), dtype=np.float64)

    elif mode in 'linear':
        schedule = np.linspace(start=temp_min, stop=temp_max, num=n_temps)

    elif mode in 'parabolic':
        temp_range = temp_max - temp_min
        temp_ind = np.arange(n_temps)
        schedule = temp_min - (temp_range / np.square((n_temps - 1.) / 2.)) * temp_ind * (temp_ind - n_temps + 1.)

    else:
        raise AttributeError(f'mode {mode} not supported')

    return schedule


if __name__ == '__main__':
    print(temp_schedule(temp_min=300, temp_max=440, n_temps=5, mode='geometric'))
    # print(temp_schedule(temp_min=310, temp_max=500, n_temps=5, mode='parabolic'))


"""
contacts.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""

from molecular.geometry import distance


# Compute the contacts between groups
def _distance_matrix(xyz0, xyz1):
    return distance(xyz0, xyz1)






"""
go.py
written in Python3
Author: C. Lockhart <chris@lockhartlab.org>

>>> go = Go(10)
>>> go.add_particle()
"""

from molecular.misc import experimental

import numpy as np


@experimental
class Go:
    """


    """

    # Initialize class instance
    def __init__(self, board, mask):
        """

        Parameters
        ----------
        shape : int, tuple with 2 elements, or numpy.ndarray
            The size of the Go board
        """

        # Shape is coerced into a 2D boolean array; True marks valid locations; False are invalid
        if isinstance(board, int):
            shape = np.ones((board, board), dtype='bool')

        else:
            raise AttributeError('board type (%s) not understood' % type(board))

        self.board = board
        self.shape = self.board.shape
        self.board = np.zeros(self.shape.shape, dtype=object)  # is this the most efficient way to do this?

    def _iterate(self):
        pass

    def _random_position(self):
        shape = self.shape.shape

    def add_stone(self, stone, position=None):
        """

        Parameters
        ----------
        stone
        position

        Returns
        -------

        """

        # If position is None, place randomly
        if position is None:
            position = self._random_position()

        # Make sure position is not unavailable
        if not self.mask[position]:
            raise AttributeError('cannot place stone in this position')

        # Make sure position is not occupied
        #

        # Put the Stone in the position
        self.board[position] = stone


class Stone:
    """
    In the game of Go, the bead things are called stones. Hence, `Stone`.
    """

    def __init__(self):
        pass




from . import analysis
from .analysis import *

from . import bioinformatics
from .bioinformatics import *

from . import core
from .core import *

from . import io
from .io import *

from . import simulations
from .simulations import *

from . import statistics
from .statistics import *

from . import transform
from .transform import *


# from . import external
# from .external import *

# from . import geometry
# from .geometry import *

# from . import viz
# from .viz import *

from .misc import *
from .version import __version__

# Contents
# __all__ = [
#     'io',
#     'include_dir',
#     'misc',
#     # 'simulation',
#     '__version__'
# ]
__all__ = ['__version__']
__all__.extend(analysis.__all__)
__all__.extend(bioinformatics.__all__)
__all__.extend(core.__all__)
# __all__.extend(external.__all__)
# __all__.extend(geometry.__all__)
__all__.extend(io.__all__)
__all__.extend(simulations.__all__)
__all__.extend(statistics.__all__)
__all__.extend(transform.__all__)
# __all__.extend(viz.__all__)

# Add include path
import os

include_dir = os.path.abspath(__file__ + '/../../include')


# logging
# TODO let people disable this, or send it to a file
import logging
logging.basicConfig(
    # filename='molecular.log',
    # filemode='w',
    format='%(asctime)s %(levelname)s (%(name)s) %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
    level=logging.INFO
)
logger = logging.getLogger('molecular')
logging.info(f'loaded version {__version__}')


"""
secondary_structure.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""

from molecular.core import Quantity, Trajectory
from molecular.external import stride
from molecular.statistics import Block

from functools import lru_cache, partial
from glovebox import GloveBox
import logging
import numpy as np
import os
import pandas as pd

logger = logging.getLogger('molecular.analysis')


def _expand_secondary_structure(data):
    """
    Expands secondary structure strings from condensed format (ex: CCCC) to expanded format (ex: C C C C, where each
    residue has its own DataFrame column).

    Parameters
    ----------
    data : pandas.Series
        Secondary structure in condensed format.

    Returns
    -------
    pandas.DataFrame
        Secondary structure in expanded format.
    """

    # This method only works on Series
    if not isinstance(data, pd.Series):
        raise AttributeError('data must be Series')

    # Convert from condensed to expanded format
    data = data.str.join(',').str.split(',', expand=True)
    data.index.name = 'structure'
    data.columns.name = 'residue'

    # Return
    return data


# Class to store secondary structure results
class SecondaryStructure:
    """
    Store the result of secondary structure computation through STRIDE.

    By itself, an instance of `SecondaryStructure` is not a `Quantity`. There are several options for converting an
    instance of `SecondaryStructure` to a `Quantity`.
    """

    # STRIDE secondary structure codes
    codes = ['H', 'G', 'I', 'E', 'B', 'b', 'T', 'C']

    # Initialize instance of SecondaryStructure
    def __init__(self, data, condensed=False):
        """
        Initialize instance of SecondaryStructure class.

        Parameters
        ----------
        data : pandas.Series or pandas.DataFrame
            The secondary structure data from STRIDE. Contains a column for each residue_id, and contains a row for
            each observation. Element represent the secondary structure code from STRIDE.
        condensed : bool
            Is `data` condensed as pandas.Series, or expanded as pandas.DataFrame? (Default: False)
        """

        # Convert condensed Series to expanded DataFrame
        if condensed:
            data = _expand_secondary_structure(data)
            logger.info(f'expanded secondary structure to shape {data.shape}')

        # At this point, we must have a DataFrame
        if not isinstance(data, pd.DataFrame):
            raise AttributeError('must be pandas DataFrame, not %s' % type(data))

        # Set instance value
        self._data = data
        self._data_long = None
        self._data_condensed = None

    def __getitem__(self, item):
        data = self._data.iloc[item, :].to_frame().T
        return SecondaryStructure(data)

    # Override __repr__
    def __repr__(self):
        # return str(self._data.agg(''.join, axis=1).values)
        # return self._condense_data()
        return str(self._data)

    def _condense_data(self):
        if self._data_condensed is None:
            self._data_condensed = self._data.agg(''.join, axis=1)
        return self._data_condensed

    # Handle codes
    @staticmethod
    def _handle_codes(codes):
        # If codes is not defined, use the global codes
        if codes is None:
            codes = SecondaryStructure.codes

        # Convert codes to numpy array for simplicity
        codes = np.array(codes).reshape(-1)

        # Make sure that all codes are permissible
        if ~np.isin(codes, SecondaryStructure.codes).all():
            bad_codes = [code for code in codes if code not in SecondaryStructure.codes]
            raise AttributeError('bad codes %s were provided' % bad_codes)

        # Return
        return codes

    # Convert from wide to long
    def _wide_to_long(self):
        """
        Convert from wide to long format.

        Returns
        -------
        pandas.DataFrame
            Secondary structure in long format.
        """

        # Convert to (residue, structure, code) format
        if self._data_long is None:
            self._data_long = self._data.stack().to_frame('code').reset_index()

        # Return
        return self._data_long

    # Count
    # TODO this could be memoized
    @lru_cache(maxsize=None)
    def count(self, axis=0):
        """
        Compute the count of secondary structure types. If axis=0, the count per residue is returned. If axis=1, the
        count per structure is returned. If count is None

        Parameters
        ----------
        axis : int or None
            (Default: 0)


        Returns
        -------
        pandas.Series or pandas.DataFrame

        """

        # Convert from wide to long
        data = self._wide_to_long()

        # axis=0 is residue
        if axis == 0:
            result = pd.crosstab(data['residue'], data['code'])

        # axis=1 is structure
        elif axis == 1:
            result = pd.crosstab(data['structure'], data['code'])

        # axis is None is count over all residues and structures
        elif axis is None:
            result = data.value_counts('code')

        # if we've made it here, something is wrong
        else:
            raise AttributeError(f'axis {axis} not understood')

        # Log
        logger.info(f'computed secondary structure count')

        # Return
        return result

    # Mean
    def mean(self, axis=0):
        """
        Compute the average secondary structure by code for the entire protein (axis = None), each structure (axis = 0),
        or each residue (axis = 1).

        Parameters
        ----------
        axis : None or int.
            Designates if the average should be computed for the entire protein (axis = None), each structure
            (axis = 0), or each residue (axis = 1). (Default: 0)

        Returns
        -------
        pandas.Series or pandas.DataFrame
            If axis = None, pandas.Series is returned. Otherwise, the result will be pandas.DataFrame.
        """

        # Get the axis counts
        data = self.count(axis=axis)

        # If axis 0 or 1
        if axis in [0, 1]:
            # noinspection PyArgumentList
            result = data.div(data.sum(axis=1), axis=0).fillna(0)

        # Otherwise, if axis is None
        else:
            result = data / data.sum()

        # Log
        logger.info(f'computed secondary structure mean')

        # Return
        return result

    # Replace
    def replace(self, old_code, new_code):
        """
        Global replacement of secondary structure type.

        Parameters
        ----------
        old_code : str
        new_code : str

        Returns
        -------

        Examples
        --------
        >>> ss = SecondaryStructure(...)
        >>> ss.replace('G', 'H')
        """

        # Log the changes
        logger.info(f'replacing secondary structure {old_code} with {new_code}')

        # Replace and store the data
        self._data = self._data.replace(old_code, new_code)

    # SEM from block averaging
    def sem_block(self, n_blocks=10):
        # Block data
        blocks = Block(self._data, n_blocks=n_blocks)

        # Convert to long format
        # noinspection PyProtectedMember
        data = blocks._data.reset_index().set_index(['structure', 'block']).stack().to_frame('code').reset_index()

        # Get counts
        counts = data.pivot_table(index=['block', 'residue'], columns='code', values='structure', aggfunc='count')\
            .fillna(0)

        # Get averages
        averages = counts.div(counts.sum(axis=1), axis=0).fillna(0.).reset_index()

        # Get std
        std = averages.pivot_table(index='residue', values=counts.columns, aggfunc=partial(np.std, ddof=1, axis=0))

        # Compute SEM
        sem = std / np.sqrt(n_blocks)

        # Log
        logger.info(f'computed secondary structure SEM with n_blocks = {n_blocks}')

        # Compute SEM
        return sem

    # Standard deviation
    def std(self, axis=None, codes=None):
        pass

    # Sum
    def sum(self, axis=None, codes=None):
        # Handle codes
        codes = self._handle_codes(codes)

        # x
        result = {}
        for code in codes:
            if axis in [0, 1]:
                result[code] = (self._data == code).sum(axis=axis)
            else:
                result[code] = (self._data == code).sum(axis=0).sum()
        return result

    # Do residues have code?
    def residues_with_code(self, code, residue_id=None):
        """
        `Quantity`

        Parameters
        ----------
        code
        residue_id

        Returns
        -------

        """

        # Handle code
        if code is None:
            raise AttributeError('code cannot be None')
        code = self._handle_codes(code)

        # Handle residue_id
        if residue_id is None:
            residue_id = self._data.columns
        residue_id = np.array(residue_id, dtype='int').reshape(-1)

        # Return as Quantity
        return Quantity(self._data[residue_id].isin(code))

    @property
    def sequence(self):
        return self._condense_data().values

    def to_csv(self, fname):
        """
        Write secondary structure object to comma-separated file.

        Parameters
        ----------
        fname

        Returns
        -------

        """

        self._data.to_csv(fname)


# Compute secondary structure using stride for a trajectory
def secondary_structure(trajectory, executable='stride', progress_bar=False):
    """
    Compute the secondary structure for :class:`molecular.Trajectory`.

    Parameters
    ----------
    trajectory : Trajectory
        An instance of Structure or Trajectory
    executable : str
        Path to stride executable. (Default: stride)
    progress_bar : bool
        Should we show a progress bar? (Default: False)

    Returns
    -------
    SecondaryStructure
        The secondary structure for the object.
    """

    # If not Trajectory, throw an error
    if not isinstance(trajectory, Trajectory):
        raise AttributeError('cannot interpret %s' % trajectory)

    # Filter only peptide atoms out of Trajectory
    peptide_trajectory = trajectory.query('peptide', only_index=False)

    # Return the result applied to the peptide as SecondaryStructure instance
    result = pd.DataFrame(peptide_trajectory.apply(partial(_compute_secondary_structure, executable=executable),
                                                   progress_bar=progress_bar))
    result['structure_id'] = np.arange(peptide_trajectory.n_structures)
    return SecondaryStructure(result.set_index('structure_id'))


# Compute the secondary structure for a single structure
def _compute_secondary_structure(structure, executable='stride'):
    """
    Compute the secondary structure for a :func:`molecular.Structure` instance

    Parameters
    ----------
    structure : Structure
        Instance of Structure class.
    executable : str
        Path to STRIDE executable.

    Returns
    -------
    pandas.Series
        The secondary structure
    """
    # Create temporary PDB in glovebox
    gb = GloveBox('molecular-stride', persist=True)
    temp_pdb = os.path.join(gb.path, 'temp.pdb')
    structure.to_pdb(temp_pdb)

    # Run STRIDE
    result = stride(temp_pdb, executable=executable)[['residue_id', 'secondary_structure']]

    # Clean out the glovebox
    gb.clean()

    # Return
    return result.set_index('residue_id').iloc[:, 0]



"""
block.py

author: C. Lockhart <chris@lockhartlab.org>
language: Python3
"""

import numpy as np
import pandas as pd
from warnings import warn


class Block:
    def __init__(self, data, n_blocks=10):
        # Make sure data is DataFrame
        if not isinstance(data, pd.DataFrame):
            raise AttributeError('data must be pandas.DataFrame')

        # Create a copy of the DataFrame
        data = data.copy()

        # Split into blocks
        if 'block' in data.columns:
            raise AttributeError('block cannot be in pandas.DataFrame')
        block_size = len(data) / n_blocks  # should this be ceiled?
        data['block'] = (np.arange(len(data)) / block_size).astype(int)

        # Make sure that we computed the right number of blocks
        if n_blocks != data['block'].nunique():
            raise ValueError('computed wrong number of blocks')

        # Save data
        self._data = data

    @property
    def data(self):
        return self._data

    # def __getattr__(self, item, *args, **kwargs):
    #     result = getattr(self._data, item, None)
    #     if result is not None:
    #         if callable(result):
    #             result = result(*args, **kwargs)
    #         if result is not None:
    #             return result

    def count(self):
        return self._data.pivot_table(index='block', aggfunc='count')

    def mean(self):
        return self._data.pivot_table(index='block', aggfunc='mean')

    def sem(self):
        # Compute standard error around mean
        return block_error(self.mean())


def block_average(df, n_blocks=10):
    return Block(df, n_blocks=n_blocks).mean()


# Shortcut function to compute block error directly from a DataFrame
def block_error(df, n_blocks=10):
    # If block is in the DataFrame, assume we're passed block averages
    if 'block' in df.columns:
        _, counts = np.unique(df['block'], return_counts=True)
        if np.max(counts) != 1:
            raise AttributeError('cannot parse DataFrame blocks')

    # Otherwise, produce the block averages
    else:
        df = Block(df, n_blocks=n_blocks).mean()

    # Return the block error
    return df.std(ddof=1) / np.sqrt(len(df))


# noinspection PyShadowingNames
def sem_block(a, indices_or_sections=10, errors='ignore'):
    r"""
    Compute the block error around the mean for a 1 or 2D array. If 2D array, the second dimension is taken as a
    collection of independent measurements.

    An array :math:`a` is divided into :math:`N` blocks such that :math:`a = \{a_1, a_2, ..., a_N\}`. For any subset
    :math:`a_i`, we can compute the average over all samples :math:`j = 1 ... M_i`,

    .. math:: \overline{a_i} = \frac{1}{N}\sum_{j=1}^{M_i}a_i(j)

    This leaves us with a set of block estimators :math:`a_{block} = \{\overline{a_1}, \overline{a_2}, ...,
    \overline{a_N}\}`. We can therefore compute the average of all blocks and the sample standard deviation,

    .. math:: \overline{a_{block}} = \frac{1}{N}\sum_{i=1}^{N}\overline{a_i}

    .. math:: s = \sqrt{\frac{1}{N-1}\sum_{i=1}^{N}(\overline{a_i}-\overline{a_{block}})^2}

    Finally, we compute the standard error,

    .. math:: Standard\ error = \frac{s}{\sqrt{N}}

    Note that the sample standard deviation is computed instead of the population standard deviation, assuming that
    :math:`N` is small.


    Parameters
    ----------
    a : numpy.ndarray
        One or two-dimensional array. First dimension is taken as independent samples. If there is a second dimension
        present, these correspond to independent measurements.
    indices_or_sections : int or list of lists
        Number of blocks or samples indices to divide into blocks.
    errors : str
        Should we `warn`, `raise`, or do nothing if the number of samples is indivisible by the number of proposed
        blocks?

    Returns
    -------
    float or numpy.ndarray
        If `a` is one-dimensional, a float will be returned that is the standard error. If two-dimensional, the result
        will be an array of standard errors, one for each independent measurement.

    References
    ----------
    [1] Flyvbjerg, H. & Petersen, H. G. (1989) Error estimates on averages of correlated data. *J. Chem. Phys.* **91**,
    461.

    [2] Frankel and Smit
      
    [3] Grossfield, A., & Zuckerman, D. M. (2009) Quantifying uncertainty and sampling quality in biomolecular
    simulations. *Annu. Rep.Comput. Chem.* **5**, 23-48.
    """

    # Number of dimensions must be 1 or 2
    if a.ndim not in [1, 2]:
        raise AttributeError('only 1 or 2 dimensional arrays allowed')

    # Warn, raise, or ignore if n_samples not evenly divisible by number of sections
    n_samples = a.shape[0]
    if isinstance(indices_or_sections, int) and np.mod(n_samples, indices_or_sections) != 0:
        message = f'{n_samples} not divisible by {indices_or_sections}'
        errors = str(errors).lower()
        if errors == 'warn':
            warn(message)
        elif errors == 'raise':
            raise AttributeError(message)
        elif errors != 'ignore':
            raise AttributeError(f'errors={errors} not understood')

    # Split
    blocks = np.array_split(a, indices_or_sections, axis=0)

    # Block averages
    blocks_avg = np.array([block.mean(axis=0) for block in blocks])

    # Compute error around average
    return np.std(blocks_avg, ddof=1, axis=0) / np.sqrt(len(blocks))

#
# if __name__ == '__main__':
#     # 1D, equal
#     a = np.random.rand(10000)
#     print(sem_block(a, 10))
#
#     # 1D, unequal
#     # a = np.random.rand(10000)
#     # print(block_error(a, 3, errors='warn'))
#
#     # 2D, equal
#     # a = np.random.rand(10000, 31)
#     # print(block_error(a))
#
#     # 2D, unequal
#     # a = np.random.rand(10000, 31)
#     # print(block_error(a, 3))



"""
trajectory.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""

from molecular.core.errata import pivot
from molecular.transform import center, move

import logging
import numpy as np
import pandas as pd
from typelike import ArrayLike, NumberLike
from tqdm import tqdm

# Get the molecular.core logger
logger = logging.getLogger('molecular.core')


# https://softwareengineering.stackexchange.com/questions/386755/sharing-docstrings-between-similar-functions
def set_doc(original):
    def wrapper(target):
        target.__doc__ = original.__doc__.replace('a : molecular.Trajectory', '')
        return target
    return wrapper


# Trajectory class
class Trajectory(object):
    """
    `Trajectory` stores instantaneous data for a `Topology`, i.e., atomic coordinates, box information, velocity, etc.
    """

    # Initialize instance of Trajectory
    def __init__(self, xyz, box=None, topology=None, copy=False):
        """


        Parameters
        ----------
        xyz : array-like
            Cartesian coordinates for all atoms and all structures. First axis must represent the structure index,
            second axis must represent the atom, and the third index will represent x, y, or z.
        topology : Topology
        """

        # If copy, make sure that xyz is decoupled from the input
        if copy or not isinstance(xyz, np.ndarray):
            xyz = np.array(xyz)

        # Extract dimensions from xyz
        n_structures, n_atoms, n_dim = xyz.shape
        if n_dim != 3:
            raise AttributeError('n_dim must be 3 for the time being')

        # Create DataFrame to store topology coordinates, velocities, force information, box information, etc.
        self._data = pd.DataFrame({
            'structure_id': np.repeat(np.arange(n_structures), n_atoms),
            'atom_id': np.tile(np.arange(n_atoms), n_structures),
            'x': xyz[:, :, 0].ravel(),
            'y': xyz[:, :, 1].ravel(),
            'z': xyz[:, :, 2].ravel()
        }).set_index(['structure_id', 'atom_id'])
        self._xyz = xyz  # TODO should this be called crd? coord? coor? pos?
        self._box = box
        self._topology = topology

    # Add
    def __add__(self, other):
        if isinstance(other, NumberLike):
            self._data[['x', 'y', 'z']] = self._data[['x', 'y', 'z']] + other
        return self

    # Length of trajectory
    def __len__(self):
        """
        The length of the trajectory is the number of structures.

        Returns
        -------
        int
            Length of trajectory
        """

        return self.n_structures

    # Get item
    def __getitem__(self, item):
        """


        Parameters
        ----------
        item

        Returns
        -------

        """

        # Return
        # TODO what if item is in topology? Should we try to retrieve that?
        # TODO should this return a DataFrame or a Trajectory object?
        return self._data.loc[item]

    # Representation of the object
    def __repr__(self):
        return "# structures: {0}\n# atoms: {1}\n# dimensions: {2}".format(*self.shape)

    # Get unique atom IDs (can be cached)
    @property
    def atom_ids(self) -> np.ndarray:
        """
        Get unique atom IDs in the Trajectory.

        Returns
        -------
        numpy.ndarray
            Array of atom IDs.
        """

        # noinspection PyUnresolvedReferences
        return self._data.index.levels[1]

    # Get coordinates
    @property
    def coord(self):
        return self.xyz

    # Set coordinates
    @coord.setter
    def coord(self, coord):
        self.xyz = coord

    # Get coordinates
    @property
    def coordinates(self):
        """
        Get Cartesian coordinates from the Trajectory.

        Returns
        -------
        pandas.DataFrame
            Trajectory Cartesian coordinates.
        """

        return self._data[['x', 'y', 'z']]

    # Set coordinates
    @coordinates.setter
    def coordinates(self, coordinates):
        """
        Set Cartesian coordinates for the Trajectory.

        Parameters
        ----------
        coordinates : array-like
            Cartesian coordinates of same shape as Trajectory.
        """

        # Set coordinates
        self._data[['x', 'y', 'z']] = coordinates

    @property
    def columns(self):
        """

        Returns
        -------
        numpy.ndarray
        """

        return self._data.columns.to_numpy()

    # Trajectory designator
    @property
    def designator(self):
        return f'Trajectory:{self.hex_id}'

    # Hex ID of Trajectory
    @property
    def hex_id(self):
        return hex(id(self))

    # Number of atoms
    @property
    def n_atoms(self):
        """
        Number of atoms in the `Trajectory`.

        Returns
        -------
        int
            Number of atoms
        """

        # noinspection PyUnresolvedReferences
        return self._data.index.levshape[1]

    # Number of dimensions
    @property
    def n_dim(self):
        """
        Number of dimensions.

        Returns
        -------
        int
            Number of dimensions
        """

        return self._data.shape[1]

    # Number of structures
    @property
    def n_structures(self):
        """
        Number of structures in the `Trajectory`.

        Returns
        -------

        int
            Number of structures
        """

        # noinspection PyUnresolvedReferences
        return self._data.index.levshape[0]

    # Shape
    @property
    def shape(self):
        """
        Shape of `Trajectory`

        Returns
        -------
        tuple
            (Number of structures, number of atoms, dimensionality)
        """

        return self.n_structures, self.n_atoms, self.n_dim

    # Get a list of structures
    @property
    def structure_ids(self) -> np.ndarray:
        """
        Get a numpy array of structure indices.

        Returns
        -------
        numpy.ndarray
            Structure indices.
        """

        # noinspection PyUnresolvedReferences
        return self._data.index.levels[0]

    # Get topology
    @property
    def topology(self):
        """

        Get the `Topology` instance.

        Returns
        -------
        Topology
            `Topology` instance associated with this `Trajectory`.
        """

        # Check the topology
        self._check_topology()

        # Return
        return self._topology

    # Get x
    @property
    def x(self):
        """
        Get x coordinates.

        Returns
        -------
        numpy.ndarray
            x coordinates.
        """

        return self._xyz[:, :, 0]

    # Get y
    @property
    def y(self):
        """
        Get y coordinates.

        Returns
        -------
        numpy.ndarray
            y coordinates.
        """

        return self._xyz[:, :, 1]

    # Get z
    @property
    def z(self):
        """
        Get z coordinates.

        Returns
        -------
        numpy.ndarray
            z coordinates.
        """

        return self._xyz[:, :, 2]

    # Get xyz coordinates
    @property
    def xyz(self):
        """
        Get x, y, and z coordinates as numpy array.

        Returns
        -------
        numpy.ndarray
            Cartesian coordinates.
        """

        return self._data[['x', 'y', 'z']].values.reshape(*self.shape)

    @xyz.setter
    def xyz(self, xyz):
        if self._xyz.shape != xyz.shape:
            raise AttributeError('must be same shape')
        self._xyz = xyz

    # Check topology
    def _check_topology(self):
        """
        Check that the topology is correctly set
        """

        # First off, must be Topology instance
        if not isinstance(self._topology, Topology):
            raise AttributeError('topology is not correct class (%s)' % type(self._topology))

        # Second, number of atoms must match
        if self.n_atoms != self._topology.n_atoms:
            raise AttributeError('number of atoms in topology and trajectory do not match ({0} vs {1})'
                                 .format(self.n_atoms, self._topology.n_atoms))

    # Apply
    # TODO please optimize this. Use vectorize, map, parallelization
    def apply(self, function, progress_bar=False):
        result = []
        structure_ids = np.arange(self.n_structures)
        if progress_bar:
            structure_ids = tqdm(structure_ids)
        for i in structure_ids:
            structure = self.get_structure(i)
            result.append(function(structure))
        return result

    # Compute the center of the Trajectory
    @set_doc(center)
    def center(self, weights=None):
        return center(self, weights=weights)

    # Copy
    # TODO
    def copy(self):
        """
        Create a copy of the Trajectory.

        Returns
        -------
        molecular.Trajectory
            Deep copy of the Trajectory.
        """

        return Trajectory(np.array(self._xyz), topology=self._topology.copy())

    # Describe
    def describe(self):
        # print('protein sequence: %s' % self.topology.residues)
        print('# structures: %s' % self.n_structures)
        print('# atoms: %s' % self.n_atoms)
        print('# dimensions: %s' % self.n_dim)

    # Get atoms
    def get_atoms(self, index):
        """
        Get specific atom indices from the Trajectory.

        Parameters
        ----------
        index : array-like
            List of atom indices.

        Returns
        -------
        numpy.ndarray
            Coordinates of specified atom indices for each structure.
        """

        # xyz will be a view if index is an int, otherwise it will be a copy...
        # TODO how will this affect behavior?
        return self._xyz[:, index, :]

    # Get column
    def get_column(self, column):
        return self._data[column]

    # Get structure
    def get_structure(self, index):
        index = np.array([index]).ravel()  # this sucks
        structure = Trajectory(self.xyz[index, :, :].reshape(len(index), self.n_atoms, self.n_dim),
                               topology=self.topology)
        if self.n_atoms != structure.n_atoms:
            raise AttributeError('number of atoms do not match ({0} vs {1})'.format(self.n_atoms, structure.n_atoms))
        return structure

    def keys(self):
        return self._topology.keys()

    # Move
    @set_doc(move)
    def move(self, by=None, to=None, inplace=True):
        return move(self, by, to, inplace)

    # Query
    def query(self, expr, only_index=False):
        """
        Query the set `Topology` and return matching indices

        Parameters
        ----------
        expr : str
            Selection text. See :ref:`pandas.DataFrame.query`
        only_index : bool
            Should the entire Trajectory be returned, or only the array of indices?

        Returns
        -------
        numpy.ndarray or Trajectory
            Array of indices or Trajectory containing only expr.
        """

        # Query the topology to get pertinent data
        topology = self.topology.query(expr)

        # Extract indices
        # noinspection PyProtectedMember
        index = topology._data.index.to_numpy()

        #
        if only_index:
            result = index

        else:
            # TODO does this work for when multiple structures should be returned?
            result = Trajectory(self.get_atoms(index).reshape(self.n_structures, len(index), self.n_dim),
                                topology=topology)

        # Return result
        return result

    # Select
    # TODO compare the efficiency of this vs query
    def select(self, **kwargs):
        """

        Parameters
        ----------


        Returns
        -------

        """

        # Save snapshot of topology data
        # noinspection DuplicatedCode,DuplicatedCode
        data = self._topology.to_frame()

        # Continuously slice the topology data
        for key in kwargs:
            item = kwargs[key]
            if not isinstance(item, ArrayLike):
                item = [item]
            data = data[data[key].isin(item)]

        # Extract indices and create a new topology
        index = data.index.to_numpy()
        topology = Topology(data)

        # Return
        # return Trajectory(self.get_atoms(index).reshape(self.n_structures, len(index), self.n_dim), topology=topology)
        return Trajectory(self.get_atoms(index), topology=topology)

    # Recenter the Trajectory
    # TODO need to decide if this should be moved to molecular.transform
    def to_center(self, weights=None, inplace=False):
        """
        Recenter all structures in the Trajectory around the origin.

        Parameters
        ----------
        weights : numpy.ndarray
            (Optional) Weights for atoms in the trajectory. Follows the definition from :ref:`numpy.average`.
        inplace : bool
            Make the change in place? (Default: False)
        """

        # Center xyz coordinates
        # noinspection PyUnresolvedReferences,PyArgumentList
        xyz = self._xyz - self.center(weights).reshape(self.n_structures, -1, self.n_dim)

        # Center Trajectory in place or return a copy
        if inplace:
            logging.info(f'centered {self.designator} at origin in place')
            self._xyz = xyz
        else:
            logging.info(f'centered {self.designator} at origin')
            trajectory = self.copy()
            trajectory._xyz = xyz
            return trajectory

    # Convert to pandas DataFrame
    def to_frame(self):
        """
        Convert `Trajectory` to pandas.DataFrame instance

        Returns
        -------
        pandas.DataFrame
            Trajectory represented as pandas DataFrame
        """

        # Prepare the data as DataFrame
        result = pd.DataFrame({
            'index': np.tile(np.arange(self.n_atoms), self.n_structures),
            'structure_id': np.repeat(np.arange(self.n_structures), self.n_atoms),
            'x': self._xyz[:, :, 0].ravel(),
            'y': self._xyz[:, :, 1].ravel(),
            'z': self._xyz[:, :, 2].ravel(),
        })

        # Return
        return result

    # Show
    def show(self):
        pass

    # Save as PDB
    def to_pdb(self, filename):
        """
        Convert `Trajectory` to PDB.

        Parameters
        ----------
        filename : str
            Name of PDB to write.
        """

        # Convert topology and trajectory to DataFrame
        topology = self.topology.to_frame()
        trajectory = self.to_frame()

        # Merge trajectory and topology
        data = trajectory.merge(topology.reset_index(), how='inner', on='index').set_index('index')

        # Sort by structure_id and then atom_id
        data = data.sort_values(['structure_id', 'atom_id'])

        # Add ATOM record
        data['record'] = 'ATOM'

        # TODO this should also be done for residue id probably
        # Change base-0 to base-1 for atom_id
        # if data['atom_id'].min() == 0:
        #     data['atom_id'] += 1

        # If we have more than 100000 atoms, we need to change atom_id
        # TODO this sort of sucks
        # https://github.com/MDAnalysis/mdanalysis/issues/1897
        fmt = '%-6s%5i %4s%4s%2s%4i%12.3f%8.3f%8.3f%6.2f%6.2f%9s%2s'
        if data['atom_id'].max() >= 100000:
            fmt = '%-6s%5s %4s%4s%2s%4i%12.3f%8.3f%8.3f%6.2f%6.2f%9s%2s'
            data['atom_id'] = pd.Series(np.vectorize(hex)(data['atom_id'])).str[2:]
            if data['atom_id'].str.len().max() >= 6:
                raise ValueError('cannot hex atom_id')

        # Format atom names
        i = data['atom'].str.len() == 1
        data.loc[i, 'atom'] = ' ' + data.loc[i, 'atom'] + '  '

        i = data['atom'].str.len() == 2
        data.loc[i, 'atom'] = ' ' + data.loc[i, 'atom'] + ' '

        # Open up the buffer
        with open(filename, 'w') as buffer:
            # Select every structure and write out
            # TODO please make this more efficient
            for structure in data['structure_id'].unique():
                is_structure = data['structure_id'] == structure

                # Select pertinent rows and columns
                structure = data.loc[is_structure, ['record', 'atom_id', 'atom', 'residue', 'chain', 'residue_id',
                                                    'x', 'y', 'z', 'alpha', 'beta', 'segment', 'element']]

                # Write out PDB file
                # TODO add box to header
                np.savetxt(
                    buffer,
                    structure,
                    fmt=fmt,
                    header='CRYST1    0.000    0.000    0.000  90.00  90.00  90.00 P 1           1',
                    footer='END\n',
                    comments=''
                )

    # Convert to MDAnalysis Universe
    # mdanalysis.org
    def to_universe(self):
        pass

    # Update
    def update(self, xyz=None):
        if xyz is not None:
            if self._xyz.shape != xyz.shape:
                raise AttributeError('must be same shape')
            self._xyz = xyz

    # View
    # https://github.com/arose/nglview
    def view(self):
        pass


# Topology class
class Topology:
    """
    A `Topology` is an object that stores metadata for a `Trajectory`, such as atomic relationships

    "Atomic" can be used loosely here because it applies to proteins, proteins in water baths, proteins in lipid
    environments, ligands in water, etc.

    The `Topology` instance generally follows PDB format [1] in that in contains,
      - atom_id
      - atom
      - residue
      - chain
      - residue_id
      - alpha
      - beta
      - segment
      - element
    """

    # Required columns
    columns = [
        'atom_id',
        'atom',
        'residue',
        'chain',
        'residue_id',
        'alpha',
        'beta',
        'segment',
        'element',
    ]

    # Initialize class instance
    def __init__(self, data=None):
        """
        Initialize class instance
        """

        # If data is set, check that it's a DataFrame
        if data is not None and not isinstance(data, pd.DataFrame):
            raise AttributeError('data must be pandas.DataFrame')

        # Add index to data
        if isinstance(data, pd.DataFrame):
            data['index'] = (data['atom_id'].rank() - 1).astype(int)
            data = data.set_index('index')

        # TODO can this be protected?
        # Save data or create blank DataFrame
        self._data = data if data is not None else pd.DataFrame(columns=self.columns)

    # Add atom to topology
    def __add__(self, other):
        pass

    # Get attribute
    def __getattr__(self, item):
        if item not in self._data:
            raise AttributeError('%s not in Structure' % item)

        return self._data[item].to_numpy()

    # Get atom from Structure by index
    def __getitem__(self, item):
        """


        Parameters
        ----------
        item

        Returns
        -------

        """

        # First, try to see if we can ping the DataFrame
        if isinstance(item, str):
            return self._data[item].to_numpy()

        # Make sure that item a valid atom_id
        if item not in self._data['atom_id']:
            raise AttributeError('%s not a valid atom_id' % item)

        # Create copy of the row
        data = self._data[self._data['atom_id'] == item].copy()

        # Sanity
        if len(data) != 1:
            raise AttributeError('multiple atom_id %s found in Structure' % item)

        # Return (as Series)
        return data.iloc[0, :]

    # Length of structure (returns number of atoms)
    def __len__(self):
        """
        Number of atoms in the Structure instance.

        Returns
        -------
        int
            Number of atoms in the Structure instance.
        """

        return self.n_atoms

    # Add new atoms f
    def add_atoms(self, **kwargs):
        pass

    # Apply
    def apply(self, function):
        return function(self)

    # TODO is this too specific? Evaluate if Structure should be more generalized
    # Compute secondary structure
    # def compute_secondary_structure(self, recompute=False, executable='stride'):
    #     # If secondary_structure is not in the Structure, or recompute is true, compute
    #     if 'secondary_structure' not in self._data or recompute:
    #         # TODO in lieu of directly porting stride to Python, how can this be sped up?
    #         # Write out PDB of structure to glovebox
    #         gb = GloveBox('molecular-stride', persist=True)
    #         temp_pdb = os.path.join(gb.path, 'temp.pdb')
    #         self.to_pdb(temp_pdb)
    #
    #         # Run STRIDE
    #         secondary_structure = stride(temp_pdb, executable=executable)[['residue_id', 'secondary_structure']]
    #
    #         # Clean glovebox
    #         gb.clean()
    #
    #         # Drop secondary_structure if it's already in self._data
    #         self._data = self._data.drop(columns='secondary_structure', errors='ignore')
    #
    #         # Merge STRIDE results with Structure to store
    #         self._data = self._data.merge(secondary_structure, how='inner', on='residue_id', validate='m:1')
    #
    #     # Return
    #     return self._data[['residue_id', 'secondary_structure']].drop_duplicates()

    # Copy
    def copy(self):
        """
        Create a copy of `Topology` instance.

        Returns
        -------
        molecular.Topology
            Copy of instance.
        """

        return Topology(self._data.copy())

    def keys(self):
        return self._data.columns.to_numpy()

    # Number of atoms
    @property
    def n_atoms(self):
        """
        Get the number of atoms in the `Topology`.

        Returns
        -------
        int
            Number of atoms
        """

        return len(self._data)

    # Pivot
    def pivot(self, index, columns=None, values=None, aggfunc='mean'):
        return pivot(self._data, index=index, columns=columns, values=values, aggfunc=aggfunc)

    # TODO should this return a Structure or Trajectory? TBD
    # TODO what if there were an intermediate object, like a Query class that knew how to handle subindexing / setting
    # Query
    def query(self, text):
        # TODO make this so it's customizable
        # Run macros
        text = text.lower().replace('peptide', 'residue in ["ALA", "ARG", "ASN", "ASP", "CYS", "GLN", "GLU", "GLY", ' +
                                    '"HIS", "HSD", "ILE", "LEU", "LYS", "MET", "PHE", "PRO", ' +
                                    '"SER", "THR", "TRP", "TYR", "VAL"]')

        # Get indices and parsed data
        indices = self._data.query(text).index.to_numpy()
        data = self._data.loc[self._data.index.isin(indices), :].copy()

        # Return
        return Topology(data)

    # TODO extend this to another trajectory
    # Compute RMSD
    def rmsd(self, structure):
        pass

    # Get residues
    @property
    def residues(self):
        """
        Get the residues of the Topology.

        Returns
        -------
        numpy.ndarray
            Array of residue names.
        """

        data = self._data[['residue_id', 'residue']].drop_duplicates()
        return data['residue'].to_numpy()

    # Write to CSV
    def to_csv(self, path, topology=False):
        pass

    # Convert to pandas DataFrame
    def to_frame(self):
        """
        Convert to pandas.DataFrame

        Returns
        -------
        pandas.DataFrame
            `Topology` as pandas.DataFrame
        """

        return self._data.copy()

    # Convert to trajectory
    def to_trajectory(self, structure_id=None):
        # Create and/or update structure_id if necessary
        if 'structure_id' not in self._data:
            structure_id = 0
        if structure_id is not None:
            self._data['structure_id'] = structure_id

        # Return Trajectory
        return Trajectory(self._data)

    # Get coordinates
    @property
    def xyz(self):
        """

        Returns
        -------

        """

        return self._data[['x', 'y', 'z']].to_numpy()


if __name__ == '__main__':
    import molecular as mol

    trj = mol.read_pdb('../tests/samples/trajectory.pdb')



"""
ising.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""


class Ising:
    def __init__(self, lattice, pbc=True, j=0., h=1.):
        self.lattice = lattice

    def energy(self):
        pass






ising = Ising()




# from molecular.analysis.peptide import compute_secondary_structure
from molecular.io import read_pdb


trj1 = read_pdb('trajectory_short.pdb')
# trj1.add_structure(read_pdb('x'))
# assert len(trj1) == 2

# secondary_structure = compute_secondary_structure(trj1, executable='C:/Users/Chris/Downloads/stride/stride.exe')
# ss = SecondaryStructure(secondary_structure.T)
# print(ss.mean(axis=None))
# print(ss.mean(axis=0))
# print(ss.mean(axis=1))


# str1 = trj1[0]
# assert pd.testing.assert_frame_equal(str1._data, trj1._data), (str1._data, trj1._data)
# str1.to_pdb('test.pdb')

# sel = trj1.query('structure_id == 0 & atom_id == 0')
# print(sel)
# trj1._data.loc[sel.index, 'record'] = ['FISH']
# sel = trj1.query('structure_id == 0 & atom_id == 0')
# print(sel)
# sel = trj1[0].query('structure_id == 0 & atom_id == 0')
# print(sel)

# sel = str1.query('structure_id == 0 & atom_id == 0')
# print(sel)
# str1._data.loc[sel.index, 'record'] = ['FISH']
# sel = str1.query('structure_id == 0 & atom_id == 0')
# print(sel)
# sel = trj1.query('structure_id == 0 & atom_id == 0')
# print(sel)


# for i in range(10):
#     str2 = read_pdb('test.pdb') + i
#     print(i, str1.rmsd(str2))

# ss = str1.get_secondary_structure(executable='C:/Users/Chris/Downloads/stride/stride.exe')
# print(ss)




"""
errata.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""

import pandas as pd
from privatize import privatize
import numpy as np


class ReservedDataFrame:
    def __init__(self, data, columns=None):
        self._data = pd.DataFrame(data, columns=columns)


# TODO since this also appears in `izzy`, assess if this should be modularized
# Extended pandas pivot tables
def pivot(df, index=None, columns=None, values=None, aggfunc=None):
    # First of, convert aggfunc to lowercase if it is a string
    if isinstance(aggfunc, str):
        aggfunc = aggfunc.lower()

    # TODO define this with Cython or C
    # Define new aggfunc
    if aggfunc == 'range':
        def _range(x):
            return np.max(x) - np.min(x)
        aggfunc = _range

    # Return call to pandas pivot_table function
    return df.pivot_table(index=index, columns=columns, values=values, aggfunc=aggfunc)




from .errata import *
from .quantity import *
# from .structure import Structure
from .trajectory import Topology, Trajectory

__all__ = [
    'pivot',
    # 'Quantity',
    # 'Structure',
    'Topology',
    'Trajectory'
]




import numpy as np
from typelike import ArrayLike


# Assert that an array is incremental
def assert_incremental(a, increment=1):
    """
    Assert that something like an array has entries that increment by a specific value.

    Parameters
    ----------
    a : ArrayLike
        Array to check if entries increment by specific value.
    increment : int
        Increment value.

    Raises
    ------
    AssertionError
        If `a` does not have entries that increment by `increment`
    """

    assert is_incremental(a, increment=increment)


# Test if an array is incremental
def is_incremental(a, increment=1):
    return (np.diff(a) == increment).all()





import molecular.flags as flags

from molecular.energy import Bond, LennardJones
from molecular.geometry import distance
from molecular.misc import experimental

import numpy as np
import pygame
from scipy.optimize import minimize as _minimize
import sys


@experimental(allow=flags.use_experimental)
class Random2D:
    def __init__(self, n_particles=100, box=(800, 600), draw=True):
        # Initialize game
        self.screen = _initialize_pygame(box)

        # Generate initial coordinates
        self.xyz = np.random.rand(n_particles, 2) * np.array(box)[None, :]

    def draw(self):
        self.screen.fill((255, 255, 255))
        for xyz in self.xyz:
            pygame.draw.circle(self.screen, (0, 0, 0), (int(xyz[0]), int(xyz[1])), 10)
        pygame.display.update()
        pygame.time.wait(100)

    def minimize(self):
        def f(xyz, screen):
            xyz = xyz.reshape(-1, 2)

            energy = 0.
            for i in range(xyz.shape[0] - 1):
                for j in range(i, xyz.shape[0]):
                    lj = LennardJones(xyz[i, :], xyz[j, :], sigma=50.)
                    energy += lj.energy

            screen.fill((255, 255, 255))
            for xyz in xyz:
                pygame.draw.circle(screen, (0, 0, 0), (int(xyz[0]), int(xyz[1])), 10)
            pygame.display.update()
            pygame.time.wait(100)

            print(energy)

            return energy

        _minimize(f, x0=self.xyz, args=self.screen)

    def molecular_dynamics(self, n_steps=1000, timestep=1e-2):
        # Run through all steps
        for step in range(n_steps):
            bond = Bond(self.xyz[0, :], self.xyz[1, :], ideal_value=0., force_constant=20.)
            force = bond.force.ravel()

            lj = LennardJones(self.xyz[0, :], self.xyz[1, :], sigma=10/np.power(2, 1/6), epsilon=500.)
            force += lj.force.ravel()
            print(bond.energy, lj.energy)

            self.xyz[0, :] -= timestep * force
            self.xyz[1, :] += timestep * force

            self.screen.fill((255, 255, 255))
            pygame.draw.circle(self.screen, (0, 0, 0), (int(self.xyz[0, 0]), int(self.xyz[0, 1])), 10)
            pygame.draw.circle(self.screen, (255, 0, 0), (int(self.xyz[1, 0]), int(self.xyz[1, 1])), 10)
            pygame.display.update()
            pygame.time.wait(100)


def _initialize_pygame(box):
    pygame.init()
    screen = pygame.display.set_mode(box)
    screen.fill((255, 255, 255))
    pygame.display.update()
    return screen


simulation = Random2D()
# simulation.draw()
# simulation.minimize()
simulation.molecular_dynamics(timestep=1e-3)




from .secondary_structure import *
from .hydrophobicity import *
from .r1N import *
from .sasa import *

__all__ = [
    'compute_hydrophobic_moment',
    'get_relative_sasa',
    'HydrophobicityScale',
    'r1N',
    'SecondaryStructure',
    'secondary_structure',
]





from fileinput import input as input_
from glob import glob as glob_
import logging
import numpy as np
import pandas as pd
import time

# Get the molecular.io logger
logger = logging.getLogger('molecular.io')


# Globular loadtxt
def loadtxt(fname, dtype=float, glob=False, verbose=False):
    """
    A refactoring of :ref:`numpy.loadtxt` that allows for globbing files.

    Parameters
    ----------
    fname : file, str, or pathlib.Path
        Name of file.
    dtype : str or object
        File type.
    glob : bool
        Does `fname` need to be globbed?
    verbose : bool
        Should information about the read-in be displayed?

    Returns
    -------
    pandas.Series
        Read file
    """

    # If glob, change fname to include all globbed files
    if glob:
        # Glob first; if glob is empty, throw an error
        fname_glob = glob_(fname)
        if not fname_glob:
            raise FileNotFoundError(fname)

        # Sort glob
        fname_glob = sorted(fname_glob)

        # Output if verbose
        if verbose:
            print(f'glob: {list(fname_glob)}')

        # Update fname to include all globbed files
        fname = input_(fname_glob)

    # Utilize numpy to read-in the file(s)
    data = np.loadtxt(fname, dtype=dtype)

    # If verbose, note the shape of the data
    if verbose:
        print(f'file loaded with shape {data.shape}')

    # Return
    return data


# TODO enable fname to be stored in the DataFrame? Is this a bad idea?
def read_table(fname, glob=False, sep='\s+', header=None, reindex=False, **kwargs):
    """
    Read table into :class:`pandas.DataFrame`.

    Parameters
    ----------
    fname : str
        Name of file.
    glob : true
        Should `fname` be treated as a globbed path? (Default: False)
    sep : str
        Character used to separate columns? (Default: white space)
    header : bool
        (Default: None)
    reindex : bool
        (Default: False)

    Returns
    -------
    pandas.DataFrame
        Data read in.
    """

    # If glob, change fname to include all globbed files
    if glob:
        # Glob first; if glob is empty, throw an error
        fname_glob = glob_(fname)
        if not fname_glob:
            raise FileNotFoundError(fname)

        # Sort glob
        fnames = sorted(fname_glob)

    # Otherwise, turn fname into a list
    # TODO evaluate if creating this list is right, or if we should short-circuit the read-in
    else:
        fnames = [fname]

    # Log files and start timer
    logger.info(f'processing file(s): {fnames}')
    start_time = time.time()

    # Cycle over fnames and read in
    kwargs['sep'] = sep
    kwargs['header'] = header
    data = [pd.read_table(_, **kwargs) for _ in fnames]

    # Concatenate
    data = data[0] if len(data) == 1 else pd.concat(data)

    # Log the shape of the data and the runtime
    end_time = time.time()
    logger.info(f'files loaded with shape {data.shape} in {int(end_time - start_time)} seconds')

    # If header is None and index_col is defined, reset columns
    if header is None and kwargs.get('index_col', None) is not None:
        data.columns = np.arange(len(data.columns))

    # Reindex?
    if reindex:
        data = data.reset_index(drop=True)

    # Return
    return data



"""
setup.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""

from numpy.distutils.core import Extension, setup

# Read version
with open('version.yml', 'r') as f:
    data = f.read().splitlines()
version_dict = dict([element.split(': ') for element in data])

# Convert the version_data to a string
version = '.'.join([str(version_dict[key]) for key in ['major', 'minor']])
if version_dict['micro'] != 0:
    version += '.' + version_dict['micro']
print(version)

# Read in requirements.txt
with open('requirements.txt', 'r') as stream:
    requirements = stream.read().splitlines()

# Long description
with open('README.rst', 'r') as stream:
    long_description = stream.read()

# First make sure numpy is installed
# _setup(install_requires=['numpy'])

# Then, install molecular
setup(
    name='molecular',
    version=version,
    author='C. Lockhart',
    author_email='chris@lockhartlab.org',
    description='A toolkit for molecular dynamics simulations',
    long_description=long_description,
    url="https://www.lockhartlab.org",
    packages=[
        'molecular',
        'molecular.analysis',
        'molecular.analysis.protein',
        'molecular.bioinformatics',
        'molecular.core',
        'molecular.external',
        'molecular.geometry',
        'molecular.io',
        'molecular.misc',
        'molecular.simulations',
        'molecular.statistics',
        'molecular.transform',
        'molecular.viz'
    ],
    # install_requires=[
    #     'glovebox',
    #     'numpy',
    #     'pandas',
    #     'privatize',
    #     'typelike',
    #     'hypothesis',
    #     'numba'
    # ],
    # include_package_data=True,
    # zip_safe=True,
    ext_modules=[
        Extension('molecular.io.fortran.read_dcd', ['molecular/io/src/read_dcd.f90'])
    ]
)



"""
sequence.py
"""


def shorten(a):
    d = {
        'ALA': 'A',
        'ARG': 'R',
        'ASN': 'N',
        'ASP': 'D',
        'CYS': 'C',
        'GLN': 'Q',
        'GLU': 'E',
        'GLY': 'G',
        'HIS': 'H',
        'ILE': 'I',
        'LEU': 'L',
        'LYS': 'K',
        'MET': 'M',
        'PHE': 'F',
        'PRO': 'P',
        'SER': 'S',
        'THR': 'T',
        'TRP': 'W',
        'TYR': 'Y',
        'VAL': 'V'
    }
    return [d[x] for x in a]


to_letter = shorten


def sequence(a):
    """

    Parameters
    ----------
    a

    Returns
    -------

    """

    # Select C-alpha atoms
    a = a.select(atom='CA')

    # Get residue name
    return a.topology['residue']



"""
r1N.py

language: Python
version: 3.7
author: C. Lockhart <chris@lockhartlab.org>
"""

from molecular.geometry import distance


# noinspection PyPep8Naming
def r1N(a):
    """

    Parameters
    ----------
    a : Trajectory

    Returns
    -------

    """

    # Extract the first and last residue_id
    residue_ids = a.topology['residue_id']
    residue_id1 = residue_ids.min()
    residue_idN = residue_ids.max()

    # Compute the center of masses
    com_id1 = a.select(residue_id=residue_id1).center()
    com_idN = a.select(residue_id=residue_idN).center()

    # Return the distances
    return distance(com_id1, com_idN)



"""
vmd.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""

import molecular.external

import os
import socket
import subprocess
import tempfile


# The class VMD starts up an instance of VMD that Python can connect
# with. We can send commands, send structures, and close VMD.
class VMD:
    """
    The class VMD starts up an instance of VMD that Python can connect with.
    You can send commands, send structures, and close VMD.
    """

    def __init__(self, source=None, port=45000,
                 executable='C:\\Program Files (x86)\\University of Illinois\\VMD\\vmd.exe'):
        """

        Parameters
        ----------
        port : int
        executable : str
        """

        self.port = int(port)
        self.executable = str(executable)

        self._process = None
        self._socket = None
        self.connected = False

        # if source is not None:
        #     self.source(source)
        # else:
        #     self.connect()

    # Close VMD and the socket
    def close(self):
        # Mark as no longer connected
        self.connected = False

        # If the VMD process is no longer available, return error
        if self._process.poll() is not None:
            raise RuntimeError('VMD already terminated')

        # Exit VMD
        self._socket.sendall(b'exit\n')

    # Connect to VMD
    def connect(self):
        # Launch VMD
        vmd_server_script = os.path.join(molecular.external.__path__[0], 'vmd_server.tcl')
        with tempfile.TemporaryFile() as file:
            self._process = subprocess.Popen([
                self.executable,
                '-e',
                vmd_server_script,
                '-args',
                str(self.port)
            ], stdout=subprocess.PIPE, stdin=file, stderr=subprocess.PIPE)

        # Set up the socket
        self._socket = socket.socket()

        # Wait for VMD to load and then connect
        # There is probably a circumstance where this results in an
        # infinite loop --> should there be a timeout?
        while not self.connected:
            try:
                self._socket.connect(('localhost', self.port))
                self.connected = True
            except ConnectionRefusedError:
                pass

        # Mark as connected
        self.connected = True

    # Load a structure into VMD
    def load(self, structure):
        # Create temporary file
        file = tempfile.NamedTemporaryFile(delete=False)
        file.close()

        # Get temporary file name
        filename = file.name.replace('\\', '/')

        # Write out structure (just PDB for now; eventually PSF/DCD?)
        structure.to_pdb(filename)

        # Execute load command on VMD
        self.execute('mol new ' + filename + ' waitfor all')

        # Delete temporary file
        os.remove(filename)

    # Execute command
    def execute(self, command):
        # If the socket isn't open, don't do anything
        if not self.connected:
            raise RuntimeError('connection already terminated')

        # If the VMD process is no longer available, return error
        if self._process.poll() is not None:
            raise RuntimeError('VMD already terminated')

        # Execute the command
        command = str(command) + '\n'  # Need newline
        command = command.replace('$', '$::')  # Need to use global namespace
        command = command.replace('[', '[uplevel #0 ')  # For global namespace
        self._socket.sendall(command.encode())

        # Receive results (should buffer be user-specified?)
        result = self._socket.recv(4096).decode('ascii').strip()

        # Split retcode from rest of results
        return_code = int(result[0])
        result = result[1:]

        # Raise error
        if return_code:
            print(command)
            print(result)
            raise RuntimeError('VMD command error')

        # Return
        if len(result) != 0:
            return result

    def source(self, filename, verbose=False):
        """
        Source from file on disk.

        Parameters
        ----------
        filename : str
            Path to file on disk
        verbose : bool
        """

        process = subprocess.Popen([self.executable, '-e', filename], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        if verbose:
            stdout, stderr = process.communicate()
            print(stdout.decode('ascii'))
            print(stderr.decode('ascii'))


def run_vmd_source(filename, verbose=False):
    VMD().source(filename, verbose=verbose)




from molecular.geometry import cos_angle
from molecular.misc import experimental
import numpy as np


# Order parameter
@experimental
def ordparam(a, b):
    """
    Compute the order parameter between two sets of vectors.

    Parameters
    ----------


    Returns
    -------

    """

    # Wrangle a and b into 3D matrices (# str by #
    a, b = _handle_a_b(a, b, paired=True)


    return 1.5 * cos_angle(a, b) ** 2. - 0.5



def _handle_a_b(a, b=None, paired=False):
    # Make sure we have numpy array
    a = np.array(a)

    # If 1D, convert to 3D
    if a.ndim == 1:
        # noinspection PyArgumentList
        a = a.reshape(1, 1, -1)

    # If 2D convert to 3D
    if a.ndim == 2:
        a = a.reshape(1, *a.shape)

    # Handle b if necessary
    if b is not None:
        # noinspection PyNoneFunctionAssignment
        b = _handle_a_b(b)
        if paired and a.shape != b.shape:
            raise AttributeError('must be same number of parameters')
        a = (a, b)

    return a
